// code for taking to the ultrasonic derived from the UltrasourndSensor tutorial
//      http://arduino.cc/en/Tutorial/UltrasoundSensor


#include "Arduino.h"
#include "osv3_dock_sensor.h"
#include "hc_sr04_ping.h"

HcSr04Ping::HcSr04Ping()
  : Osv3Thgr810Sensor(HEIGHT_CHANNEL, TX_DATA_PIN, TX_POWER_PIN)
  , m_waterLevel(15)     // 15 minutes of data to average
  , m_waveHeights(m_secondsBetweenTransmits[HEIGHT_CHANNEL - 1])
{
  // configure pins
  pinMode(PING_TRIGGER_PIN, OUTPUT);
  pinMode(PING_POWER_PIN, OUTPUT);
  pinMode(ECHO_PIN, INPUT);  

  // always leaving power on
  //digitalWrite(PING_POWER_PIN, LOW);
  digitalWrite(PING_POWER_PIN, HIGH);  
  
  // first few readings are garbage
  readDistance();
  readDistance();
  readDistance();
  readDistance();
}

HcSr04Ping::~HcSr04Ping()
{
}

void HcSr04Ping::loop(void)
{
  // called once per second
  
  // want to display water heigh in inches (cm may be better choice for others)
  // temp is displayed in F (C may be better for others)
  // 0.1 change in F will represent 1/10 of an inch
  float distance;
  
  // samples before sending first packet, get a few for reasonable high/low values
  static unsigned int samplesUntilTransmit = 15;

  static unsigned long txTiming = millis();

  samplesUntilTransmit--;

  // ping returns distance in 1/10 of an inch
  distance = readDistance(5);
  
  if (distance > 0)
  {
     m_waveHeights.store(distance);
  }

  Serial.print(samplesUntilTransmit + 1);
  Serial.print(") Distance (tenths): ");
  Serial.println(distance); 
  
  // send every 59 seconds
  if (samplesUntilTransmit <= 0)
  {
    digitalWrite(LED_PIN, HIGH);
    
    Serial.print("height tx timing: ");
    Serial.println(millis() - txTiming);
    txTiming = millis();
    
    m_waveHeights.removeOutliers();
    
    buildAndSendHeightPacket(m_waveHeights.minimum(), m_waveHeights.maximum());

    m_waveHeights.clear();
    
    digitalWrite(LED_PIN, LOW);

    samplesUntilTransmit = m_secondsBetweenTransmits[HEIGHT_CHANNEL - 1];
  } 
}

float HcSr04Ping::readDistance(int attempts)
{ 
  float reading;
  
  // will retry read if distance comes back as 0
  for (;attempts>0; attempts--)
  {
     reading = readDistance();
     if (reading != 0.0) 
     {
       return reading;
     }
     delay(40);
  }
  
  return 0.0;
}

float HcSr04Ping::readDistance(void)
{
  unsigned long duration;
  float distance;

  // not working?  bad sensor?
  //digitalWrite(PING_POWER_PIN, HIGH);
  //delay(1000);

  // The PING))) is triggered by a HIGH pulse of 2 or more microseconds.
  // Give a short LOW pulse beforehand to ensure a clean HIGH pulse:
  digitalWrite(PING_TRIGGER_PIN, LOW);
  delayMicroseconds(2);
  digitalWrite(PING_TRIGGER_PIN, HIGH);
  delayMicroseconds(10);
  digitalWrite(PING_TRIGGER_PIN, LOW);

  // The same pin is used to read the signal from the PING))): a HIGH
  // pulse whose duration is the time (in microseconds) from the sending
  // of the ping to the reception of its echo off of an object.
  duration = pulseIn(ECHO_PIN, HIGH, 50000);

  // convert the time into a distance
  distance = microsecondsToInches(duration);

  //digitalWrite(PING_POWER_PIN, LOW);
  return distance;
}

float HcSr04Ping::microsecondsToInches(unsigned long microseconds)
{
  // According to Parallax's datasheet for the PING))), there are
  // 73.746 microseconds per inch (i.e. sound travels at 1130 feet per
  // second).  This gives the distance travelled by the ping, outbound
  // and return, so we divide by 2 to get the distance of the obstacle.
  return microseconds / 147.492;
}

void HcSr04Ping::buildAndSendHeightPacket(float low, float high)
{
  float waveHeight = high - low;
  unsigned int level;
  
  // no measurement made
  if (high == 0.0)
  {
    Serial.println("No readings, sending 0 height and avg level");
    buildAndSendPacket(-17.777, m_waterLevel.average());
    return;
  }
  
  // we want the display of 1F to be 1", but data is sent as C, so we have to convert
  // a 74.3" wave would be represented as 74.3F
  // this is converted to 23.5C, and represent it here as 235
  waveHeight = ((waveHeight - 32) * 5)  / 9.0;
  
  Serial.println("Send packet");
  Serial.print("high: ");
  Serial.print(high);
  Serial.print("; low: ");
  Serial.println(low);
  Serial.print("waves: ");
  Serial.println(waveHeight);

  // using humidity field for low water level in inches 0-100"
  // the high distance measurement is actually the low water mark
  // I want to send a water level as a delta from some point
  // in this case, I'll just use 100" as sensor height which puts 0" at 8' 4" down from the dock
  // as my water level should never be lower than that

  // average data height over 15 minutes  
  m_waterLevel.store(99 - ((high + low) / 2));
  level = m_waterLevel.average();
  Serial.print("height: ");
  Serial.println(level);

  // send high value
  buildAndSendPacket(waveHeight, level);
}

