#ifndef __REINGBUFFER_H_
#define __REINGBUFFER_H_

// simple ring buffer class used to average sensor samples
// uses templates so various simple data types (long, float) can be stored and averaged

#include <Arduino.h>
#include <stdlib.h>

template<typename kind>
class RingBuffer
{
   private:
      kind *buffer;
      int maxSamples;
      int writeLocation;
      int numSamplesStored;
      
      RingBuffer();

     void removeSample(int index)
     {
         int i;
         for (i=index+1; i<numSamplesStored; i++)
         {
            buffer[i-1] = buffer[i];
         }
      
         numSamplesStored--;
     }
     
      kind maximumIndex(void)
      {
         int i;
         int index = 0;
      
         for (i=1; i<numSamplesStored; i++)
         {
            if (buffer[i] > buffer[index])
            {
              index = i;
            }
         }
      
         return index;
     }     
     
      kind minimumIndex(void)
      {
         int i;
         int index = 0;
      
         for (i=1; i<numSamplesStored; i++)
         {
            if (buffer[i] < buffer[index])
            {
              index = i;
            }
         }
      
         return index;
     }     
  
     
     // estimate of Grubbs critical value for 95%
     float grubbsCriticalValue(void)
     {
       return ((3.6996 * numSamplesStored) + 145.9 - (186.7 / numSamplesStored)) / (numSamplesStored + 59.5 + (58.5 / numSamplesStored));
       
       // for 99% use this equation to instead
       // return ((4.1068 * numSamplesStored) + 273.6 - (328.5 / numSamplesStored)) / (numSamplesStored + 88.7 + (185 / numSamplesStored));
     }

  
  public:

      RingBuffer(int maxSamplesRequested)
      {
         clear();
      
         // no new/delete functions available
         buffer = (kind*)malloc(sizeof(kind) * maxSamplesRequested);
         if (buffer == NULL)
         {
           // malloc failed
           maxSamples = 0;
         }
         else
         {
           maxSamples = maxSamplesRequested;
         }
      }
      
      ~RingBuffer()
      {
        if (buffer != NULL)
        {
          free(buffer);
        }
      }

      void clear(void)
      {
         writeLocation = 0;
         numSamplesStored = 0;        
      }
      
      void store(kind x)
      {
         buffer[writeLocation] = x;
      
         if (++writeLocation >= maxSamples)
         {
            writeLocation = 0;
         }
      
         if (++numSamplesStored >= maxSamples)
         {
            numSamplesStored = maxSamples;
         }
      }
      
      kind average(void)
      {
         kind sum = 0;
         int i;
      
         for (i=0; i<numSamplesStored; i++)
         {
            sum += buffer[i];
         }
      
         return sum/numSamplesStored;
      }
     
      float standardDeviation(void)
      {
         kind avg = average();
         kind diff;
         kind sum = 0;
         int i;
      
         for (i=0; i<numSamplesStored; i++)
         {
           sum += pow((buffer[i] - avg), 2);
         }
      
         return sqrt(sum/(float)numSamplesStored);
      }     

      kind maximum(void)
      {
         kind x = buffer[0];
         int i;
      
         for (i=1; i<numSamplesStored; i++)
         {
            x = max(x, buffer[i]);
         }
      
         return x;
     }     
     
      kind minimum(void)
      {
         kind x = buffer[0];
         int i;
      
         for (i=1; i<numSamplesStored; i++)
         {
            x = min(x, buffer[i]);
         }
      
         return x;
     }          
     
     // Grubbs test for outliers
     void removeOutliers(void)
     {
       bool removed = false;
       
       do 
       {
         kind avg = average();
         int maxIndex = maximumIndex();      
         int minIndex = minimumIndex();      
         float stdDev = standardDeviation();
         float z;
         float criticalValue = grubbsCriticalValue();
       
         removed = false;
       
         Serial.print("mean: ");
         Serial.print(avg);
         Serial.print("; std dev: ");
         Serial.print(stdDev);
         Serial.print("; critical value: ");
         Serial.println(criticalValue);
         
         z = (buffer[maxIndex] - avg) / stdDev;
         if (z > criticalValue)
         {
           Serial.print("max outlier: ");
           Serial.print(buffer[maxIndex]);
           Serial.print("; z: ");
           Serial.println(z);
           removeSample(maxIndex);
           removed = true;
         }

         z = (avg - buffer[minIndex]) / stdDev;
         if (z > criticalValue)
         {
           Serial.print("test min outlier: ");
           Serial.print(buffer[minIndex]);
           Serial.print("; z: ");
           Serial.println(z);
           removeSample(minIndex);
           removed = true;
         }
       }  while (removed);
     }
};


#endif
